classdef Communicator < handle
    %MESSAGEGENERATOR Summary of this class goes here
    %   Detailed explanation goes here
    
    properties
        nodeId;
        fs;
        fc;
        fm;
        l_signal;
        code;
        n_code;
        
        n_node;
        
        spacing = 0.22; % Spacing between microphones
        c_sound = 343;  % Speed of sound
        max_delay = 600E-6; % Maximum delay between mics
        
        % Code 1:3:N-2  is Requests
        % Code 2:3:N-1  is Response
        % Code 3:3:N    is Retransmissions
        
        % Calibration
        update_time = 2000;    % Update frequency in milliseconds
        processing_delay = 211569;
        calInProgress = 0;
        
        pending_request_out;    % Describes pending requests at lag, otherwise 0
        pending_response_out;   % Describes pending requests at lag, otherwise 0
        pending_retrans_out;    % Describes pending answers at lag, otherwise 0
        pending_request_in;     % Describes listens as 1, otherwise 0
        pending_response_in;    % Describes listens as 1, otherwise 0
        pending_retrans_in;     % Describes listens as 1, otherwise 0
        last_update;            % Last update for node
        fixed_delays;           % Fixed delays for node (in signal lengths
        queue;                  % buffer number to start transmission of signal
        
        current_lag;
        
        modulator;
        
        distance;
        
        
    end
    
    methods
        function self = Communicator(nodeId, fs, fc, fm, l_signal, code, n_code, modulator)
            self.nodeId = nodeId;
            self.fs = fs;
            self.fc = fc;
            self.fm = fm;
            self.l_signal = l_signal;
            self.code = code;
            self.n_code = n_code;
            
            self.n_node = ceil(self.n_code/3);
           
            self.fixed_delays = ones(1,self.n_node) * 2;
            self.pending_request_out = zeros(1,self.n_node);  
            self.pending_response_out = zeros(1,self.n_node);  
            self.pending_retrans_out = zeros(1,self.n_node);  
            self.pending_request_in = zeros(1,self.n_node);  
            self.pending_response_in = zeros(1,self.n_node);  
            self.pending_retrans_in = zeros(1,self.n_node);  
            self.last_update = zeros(2,self.n_node);
            
            self.queue = zeros(1,self.n_node) - 1;
            
            self.modulator = modulator;
            
        end
        
        function positions = step(self, updates, current_lag)
            % Takes updates as
            % Channel, Code, Total lag
            % for each lag and 
            % current_time as total lag.
            %
            % Returns positions as
            % NodeId, distance, angle
            % for each node
            
            self.current_lag = current_lag;
            
            node_update = zeros(self.n_node,2); % Updated
            positions = [];
            %disp(updates)
            if ~isempty(updates) % Incoming updates
                % For each update
                
                for i = 1:size(updates,1)
                    code_type = mod(updates(i,2)-1,3);
                    node_id  = 1 + (updates(i,2)-1 - code_type)/3;
                    ch = updates(i,1);
                    lag = updates(i,3);
                    % code_type 0 is Request
                    % code_type 1 is Response
                    % code_type 2 is Retransmission
                    % code_type -1 means no pending output
                    switch code_type
                        case 0
                            % Check if recently updated
                            if self.last_update(ch,node_id) < self.update_time*lag + current_lag
                                if node_id ~= self.nodeId
                                    self.pending_response_out(self.nodeId) = lag + self.fixed_delays(self.nodeId)*self.l_signal;
                                    self.queue(self.nodeId) = 1;
                                    fprintf('Received Node %i request\n', node_id);
                                end
                            end
                            
                        case 1
                            % Check if waiting for response
                            if self.pending_response_in(node_id) > 0
                                %self.pending_response_in(node_id) = self.pending_response_in(node_id) -  1; 
                                self.pending_retrans_out(node_id) = lag + self.fixed_delays(node_id)*self.l_signal; % Insert pending retrans
                                self.queue(node_id) = 2;
                                
                                node_update(node_id,1) = 1;
                                node_update(node_id,2) = code_type;
                                self.last_update(node_id,ch) = lag;
                                
 
                                % Calibration
                                if self.calInProgress
                                   self.calInProgress = 0;
                                   self.processing_delay = lag - self.pending_request_out(self.nodeId) - round(2/self.c_sound*self.fs);
                                   fprintf('Calibrated at %i lags.\n',self.processing_delay); 
                                end
                            end
                                
                            
                        case 2
                            % Check if waiting for retrans
                            if self.pending_retrans_in(node_id) > 0
                                self.pending_retrans_in(node_id) = self.pending_retrans_in(node_id) - 1; % Clear pending retrans
                                 
                                node_update(node_id,1) = 1;
                                node_update(node_id,2) = code_type;
                                self.last_update(node_id,ch) = lag;
                                
                                
                            end
                            
                        otherwise
                            fprintf('Communicator step function error\n');
                    end
                end
                
                
                % Calculate distances and angles
                pos_idx = 1;
                for i = 1:self.n_node
                    if node_update(i,2) > 0
                        positions(pos_idx,1) = i;
                        lag = self.last_update(i,:);
                        if abs(lag(1) - lag(2)) < 10000
                            mean_lag = sum(lag)/2;                        
                            angle1 = 180/pi * asin(self.c_sound*(lag(1)-lag(2))/self.fs/self.spacing);
                            angle2 = self.c_sound*(lag(1)-lag(2))/self.fs/self.spacing/pi*180;
                            angle= (angle1+angle2)/2.02;
                        else
                            [mean_lag, detected_ch] = max(lag);
                            angle = -270 + detected_ch*180;
                        end
                        switch node_update(i,2)
                            case 1
                                out_lag = self.pending_request_out(self.nodeId);
                            case 2
                                out_lag = self.pending_response_out(node_id);
                        end
                        travel_lags = mean_lag - out_lag - self.processing_delay;
                        travel_time = travel_lags/self.fs;
                        travel_distance = travel_time*self.c_sound/2;
                        positions(pos_idx,2) = travel_distance;
                        positions(pos_idx,3) = angle;
                        self.distance(end+1) = travel_distance;
                        pos_idx = pos_idx + 1;
                    end
                   
                    
                end
            end % Incoming updates logic
            
            % Scan ouput queue
            for i = 1:self.n_node
                switch self.queue(i)
                    case 0
                        % Check for pending request
                        if ( self.pending_request_out(i) > current_lag ) && ( self.pending_request_out(i) <= current_lag + self.l_signal )
                            self.queue(i) = -1;
                            relative_lag = self.pending_request_out(i) - current_lag;
                            %fprintf('Sending request, node %i; Current lag %i, new lag %i ', i, current_lag, self.pending_request_out(i));
                            self.modulator.setOutput(3*i-2, relative_lag);
                            self.pending_response_in = 2*ones(1,self.n_node);
                            self.pending_response_in(self.nodeId) = 0;
                            
                        end
                
                    case 1
                        
                        % Check for pending response
                        if ( self.pending_response_out(i) > current_lag ) && ( self.pending_response_out(i) <= current_lag + self.l_signal )
                            self.queue(i) = -1;
                            relative_lag = self.pending_response_out(i) - current_lag;
                            %fprintf('Sending response, node %i; Current lag %i, new lag %i ',i, current_lag, self.pending_response_out(i));
                            self.modulator.setOutput(3*i-1, relative_lag);
                            % When sent response, wait for retransmission.
                            self.pending_retrans_in(self.nodeId) = 2;
                        end
                        
                        
                    case 2
                         % Check for pending retransmission
                        if ( self.pending_retrans_out(i) > current_lag ) && ( self.pending_retrans_out(i) <= current_lag + self.l_signal )
                            self.queue(i) = -1;
                            %fprintf('Sending retransmission, node %i; Current lag %i, new lag %i ',i, current_lag, self.pending_retrans_out(i));
                            relative_lag = self.pending_retrans_out(i) - current_lag;
                            self.modulator.setOutput(3*i, relative_lag);
                            
                            
                        end
                end
               
                
                
                
            end
            
        end
        
        
        function sendRequest(self)
            self.pending_request_out(self.nodeId) = self.current_lag + self.l_signal +1;
            self.queue(self.nodeId) = 0;
            fprintf('Comunicator.sendRequest, ID = %i curr_lag = %i, new lag = %i\n', self.nodeId, self.current_lag, self.pending_request_out(self.nodeId) );
        end
        
        function calibrate(self)
            self.calInProgress = 1;
            self.sendRequest();
        end
    end
    
end

