% RealTimeAudioProcessor
%
% The RealTimeAudioProcessor object allows the user to process consecutive
% frames of audio in real-time (it lags real-time by the frame size plus 
% hardware overhead). The object can easily be extended for any type of
% audio processing by overloading the stepImpl method in a derived class.
%
% Use:
%
% % Create the processor with a number of input channels, number of output
% % channels, sample rate, frame size, and device name.
% self = RealTimeAudioProcessor(1, 2, 48000, 256, 'Default');
% 
% % Play. This will start *immediately* and block until all audio output is
% % complete.
% self.Play();
% 
% % Release the audio resource for others (be nice).
% self.release();
% 
% % The RTAP records some timing values while playing. Show the results.
% self.Analyze();
% 
% It is recommended that an ASIO audio driver with appropriate hardware be
% used for the audio interface. (ASIO is an open standard from Steinberg
% Media Technologies GmbH).
%
% Tucker McClure @ The MathWorksaudiorecorderkkjl
% Copyright 2012 The MathWorks, Inc.
% Modified by Erlend Fasting @ NTNU 2015

classdef RTAP < matlab.System

    properties %(Nontunable, Access = protected)
        
        % Settings
        config;
        l_code = 511;
        
        
        frame_size;   % Number of samples in the buffer        
        end_time     = inf;   % Number of seconds until playback stops [s]
        in_channels  = 2;     % Number of input channels
        out_channels = 2;     % Number of output channels
        device_name;
        draw_rate    = 0.01;  % Rate to update graphics [s]
        
        running = false;
        
        % Derived quantities
        time_step;   % Length of frame [s]
        time_window; % Array of time values from [0 time_step) [s]s
        
        % Device interfaces
        ap; % AudioPlayer object to manage output
        ar; % AudioRecorder object to manage input
        
        deviceInfoIn;
        deviceInfoOut;
        
        % UI handles
        figure_handle;
        config_panel;
        runtime_panel;
        device_input_list;
        device_output_list;
        nNodes_box;
        nodeId_list;
        fs_list;
        fc_box;
        fm_box;
        baseband_list;
        status_text;
        start_btn;
        stop_btn;
        angle_box;
        set_angle_btn;
        request_btn;
        cal_btn;
        start_transmission_btn;
        stop_transmission_btn;
        set_att_btn;
        att_box;
        start_cal_btn;
        stop_cal_btn
        
        current_angle_txt;
        
        samples_until_draw = 0; % Samples left before updating GUI
        
        % Analysis stuff
        max_in_step_time      = 0;
        max_process_step_time = 0;
        max_out_step_time     = 0;
        
        
        % Objects
        node;
        mobdev = [];
        handles;
        
        turntable;
        
        out_factor = 1;
        
    end
    
    properties
        
        
        
    end
    
    methods
        
        function self = RTAP()
            
            try
                self.loadConfig();
            catch
                self.createConfig();
            end
               
            
           
            % Find available devices
            self.deviceInfoIn = dspAudioDeviceInfo('inputs');
            self.deviceInfoOut = dspAudioDeviceInfo('outputs');
            
            self.GenerateFigure();
            
            try
                self.turntable = Turntable();
            catch
                fprintf('No turntable available\n');
            end
        end
        
        
        function setupNode(self)
            
            self.printStatus('Init audio...');
             % Buffer overlap 
            self.frame_size = round(self.l_code*2*self.config.fs/self.config.fm);
            % Calculate the period.
            self.time_step = self.frame_size/self.config.fs;
            % Create all the time values for a window.
            self.time_window = (0:self.frame_size-1)'/self.config.fs;

        
           
            
                % Create the AudioPlayer.
            if self.out_channels > 0
                if strfind(self.config.deviceOutName{1},' (Core Audio)') > 0
                    name = self.config.deviceOutName{1}(1:strfind(self.config.deviceOutName{1},' (Core Audio)')-1);
                else 
                    name = self.config.deviceOutName{1};
                end
                
                self.ap = dsp.AudioPlayer(...
                    'DeviceName',       name, ...
                    'BufferSizeSource', 'Property', ...
                    'BufferSize',       self.frame_size, ...
                    'QueueDuration',    0, ...
                    'SampleRate',       self.config.fs,...
                    'OutputNumUnderrunSamples',true);
                
                
                % Start with silence. This initializes the AudioPlayer to
                % the window size and number of channels and takes longer
                % than any subsequent call will take.
                %step(self.ap, zeros(self.frame_size, self.out_channels));

            end
            
            % Create the AudioRecorder (if requested).
            if self.in_channels > 0
                if strfind(self.config.deviceInName{1},' (Core Audio)') > 0
                    name = self.config.deviceInName{1}(1:strfind(self.config.deviceInName{1},' (Core Audio)')-1);
                else
                    name = self.config.deviceInName{1};
                end
                self.ar = dsp.AudioRecorder(...
                    'DeviceName',       name, ...
                    'SampleRate',       self.config.fs, ...
                    'BufferSizeSource', 'Property', ...
                    'BufferSize',       self.frame_size, ...
                    'SamplesPerFrame',  self.frame_size, ...
                    'QueueDuration',    0, ...
                    'OutputDataType',   'double', ...
                    'NumChannels',      self.in_channels,...
                    'OutputNumOverrunSamples',true);
                
                % Initialize the input.
                %step(self.ar);
            end
            
             % Display latency to user.
            self.printStatus(sprintf('Minimum latency: %5.1fms', ...
                    1000 * self.time_step));
            
    
            % Draw it.
            drawnow();
            
            
            % Chill out for a second before rushing forward with sound.
            pause(self.time_step);
            
            
            self.printStatus('Audio init completed');
            self.printStatus('Init node...');
            drawnow();
            
            self.handles = struct(...
                'mobdev', self.mobdev,...
                'current_angle', self.current_angle_txt);
            % Create node
            self.node = Node(self.config.nodeId, self.config.nNodes, self.config.fs, self.config.fc, self.config.fm, self.handles);
            self.printStatus('Node initialization completed');
        end
        
        
        
        % Enter a quasi-real-time loop in which audio is acquired/generated
        % and plugged into the output buffer (if a buffer exists).
        function Play(self)
            
            
            % Keep track of time since 'tic'.
            t_clock = 0;
            
            % Keep track of how much material we've played since 'tic'.
            % At t_clock, this should reach to t_clock + time_step.
            t_played = 0;
            
            % Initialize the input.
            %in = zeros(self.frame_size, self.in_channels); 
            % Start a timer.
            tic();
            

            % Loop until the end time has been reached or the figure has
            % been closed.
            self.printStatus('Real-time loop started!');
            while t_clock < self.end_time && self.running
                
                % Play steps until we're |buffer| into the future.
                if t_played < t_clock + self.time_step

                    % Create the times for this frame.
                    time = t_played + self.time_window;
                    
                    
                    
                    % Get the input for one frame.
                    if self.in_channels > 0
                        step_timer = tic();
                        
                        [in, overrun] = step(self.ar);
                        if overrun > 0
                            fprintf('Overrun in samples: %i\n', overrun);
                        end
                        
                        self.max_in_step_time = ...
                            max(self.max_in_step_time, toc(step_timer));
                        
                        
                    else
                        in = zeros(self.frame_size, self.in_channels);
                    end
                    
                    % Process one frame.
                    step_timer = tic();
                    out = self.out_factor*self.node.step(in')';
                    out = [out out];
                    %out = zeros(self.frame_size,self.out_channels);
                    self.max_process_step_time = ...
                        max(self.max_process_step_time, toc(step_timer));
                    
                    % Step the AudioPlayer. Time the step for analysis
                    % purposes.
                    if self.out_channels > 0
                        step_timer = tic();
                        underrun = step(self.ap, out);
                        if underrun > 0
                            fprintf('Underrun out samples: %i \n', underrun);
                        end
                        self.max_out_step_time = ...
                            max(self.max_out_step_time, toc(step_timer));
                    end
                    
                    % Update the time.
                    t_played = t_played + self.time_step;
                    
                    
                end 
                
                % Release focus so that figure callbacks can occur.
                if self.samples_until_draw <= 0

                    drawnow;
                    self.UpdateGraphics();
                    self.samples_until_draw = ...
                        self.config.fs * self.draw_rate;
                else
                    self.samples_until_draw = ...
                        self.samples_until_draw - self.frame_size;
                end
                
                % Update the clock.
                t_clock = toc();
                
            end % end while
            
            % Wait for audio to end before exiting. We may have just
            % written out a frame, and there may have already been a frame
            % in the buffer, so chill for 2 frames.
            pause(2*self.time_step);
            self.printStatus('Real-time loop ended');
            self.Analyze();
            self.release();
        end
        
        % Display timing results from last play.
        function Analyze(self)
            fprintf(['Results for last play:\n', ...
                     'Maximum input step time:   %5.1fms\n', ...
                     'Maximum process step time: %5.1fms\n', ...
                     'Maximum output step time:  %5.1fms\n'], ...
                    1000*self.max_in_step_time, ...  
                    1000*self.max_process_step_time, ...  
                    1000*self.max_out_step_time);
        end
        
        
        function setMobdev(self, mobdev)
            
            self.mobdev = mobdev;
            
        end
        
        
    end
    
    methods (Access = protected)
 
        
        % Process one frame of audio, given the inputs corresponding to the
        % frame and the times.
        function out = stepImpl(self, time, in) %#ok<INUSD>
            out = zeros(self.frame_size, self.out_channels);
        end
        
        % Specify that the step requires 2 inputs.
        function n = getNumInputsImpl(~)
            n = 2;
        end
        
        % Specify that the step requires 1 output.
        function n = getNumOutputsImpl(~)
            n = 1;
        end
        
        % Clean up the AudioPlayer.
        function releaseImpl(self)
            
            % Release the dsp.AudioPlayer resource.
            if self.out_channels > 0
                release(self.ap);
            end
            
            % Release the dsp.AudioRecorder too.
            if self.in_channels > 0
                release(self.ar);
            end
            
            
        end
        
        % Generate a figure that stays open with updates for the user.
        % Closing this figure ends playback.
        function GenerateFigure(self)
            
            % Get the screen dimensions for centering the figure.
            figure_width_height = [640 640];
            figure_dims = [1608, 1050, figure_width_height];
                       
            % Generate a figure.
            self.figure_handle = figure(...
                'Name',          'Real-Time Audio Processor Controller',...
                'NumberTitle',   'off', ...
                'Position',      figure_dims);
            
            self.config_panel = uipanel( self.figure_handle,...
                'Title', 'Configuration',...
                'Position', [.05 .05 .4 .90]);
            
            
            self.device_input_list = uicontrol( self.config_panel,...
                'Style', 'popupmenu',...
                'String', {self.deviceInfoIn.name},...
                'Units', 'normalized',...
                'Position', [.1 .94 .8 .04]);
             
            self.device_output_list = uicontrol( self.config_panel,...
                'Style', 'popupmenu',...
                'String', {self.deviceInfoOut.name},...
                'Units', 'normalized',...
                'Position', [.1 .89 .8 .04]);
            
            nNodes_text = uicontrol( self.config_panel,...
                'Style', 'text',...
                'String', 'Number of nodes:',...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.12 .82 .4 .04]);
            
            self.nNodes_box = uicontrol( self.config_panel,...
                'Style', 'edit',...
                'String', num2str(self.config.nNodes),...
                'Units', 'normalized',...
                'Position', [.12 .79 .4 .04],...
                'Callback', @self.nNodes_box_Callback);
               
            nodeId_text = uicontrol( self.config_panel,...
                'Style', 'text',...
                'String', 'This node:',...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.62 .82 .30 .04]);
            
            self.nodeId_list = uicontrol( self.config_panel,...
                'Style', 'popupmenu',...
                'String', '1',...
                'Units', 'normalized',...
                'Position', [.6 .79 .3 .04]);
            
            static_fs_text =  uicontrol( self.config_panel,...
                'Style', 'text',...
                'String', 'Sample rate [Hz]:',...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.13 .73 .75 .04]);
            
            self.fs_list = uicontrol( self.config_panel,...
                'Style', 'popupmenu',...
                'String', {'44100' '48000' '96000' '192000'},...
                'Units', 'normalized',...
                'Position', [.1 .70 .8 .04]);
            
            static_fc_txt =  uicontrol( self.config_panel,...
                'Style', 'text',...
                'String', 'Carrier rate [Hz]:',...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.13 .63 .75 .04]);
            
            self.fc_box = uicontrol( self.config_panel,...
                'Style', 'edit',... 
                'Units', 'normalized',...
                'HorizontalAlignment', 'left',...
                'Position', [.12 .60 .75 .04]);
            
            static_fm_txt =  uicontrol( self.config_panel,...
                'Style', 'text',...
                'String', 'Message rate [Hz]:',...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.13 .54 .75 .04]);
            
            self.fm_box = uicontrol( self.config_panel,...
                'Style', 'edit',... 
                'Units', 'normalized',...
                'HorizontalAlignment', 'left',....
                'Position', [.12 .51 .75 .04]);
            
            self.start_btn = uicontrol( self.config_panel,...
                'Style', 'pushbutton',...
                'String', 'Start RealTime',...
                'Units', 'normalized',...
                'Position', [.05 .01 .45 .1],...
                'Callback', @self.start_btn_Callback);
                
            self.stop_btn = uicontrol( self.config_panel,...
                'Style', 'pushbutton',...
                'String', 'Stop RealTime',...
                'Units', 'normalized',...
                'Position', [.5 .01 .45 .1],...
                'Callback', @self.stop_btn_Callback);
            
            
            static_baseband_txt =  uicontrol( self.config_panel,...
                'Style', 'text',...
                'String', 'Use I-Q demodulation:',...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.13 .44 .75 .04]);
            
            self.baseband_list = uicontrol( self.config_panel,...
                'Style', 'popupmenu',...
                'String', {'No', 'Yes'},...
                'HorizontalAlignment', 'left',...
                'Units', 'normalized',...
                'Position', [.1 .40 .8 .04]);
            
             self.status_text = uicontrol( self.config_panel,...
                'Style', 'edit',...
                'HorizontalAlignment', 'left',....
                'MAX', 1000,...
                'Units', 'normalized',...
                'Position', [.1 .15 .8 .2]);
            
            self.runtime_panel = uipanel( self.figure_handle,...
                'Title', 'Runtime',...
                'Position', [.5 .05 .4 .90]);
            
            
            self.request_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Send Request',...
                'Units', 'normalized',...
                'Position', [.05 .01 .45 .1],...
                'Callback', @self.request_btn_Callback);
            
            self.cal_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Calibrate',...
                'Units', 'normalized',...
                'Position', [.5 .01 .45 .1],...
                'Callback', @self.cal_btn_Callback);
                                    
            self.angle_box = uicontrol(self.runtime_panel, ...
                'Style', 'edit',...
                'String', '0',...
                'Units', 'normalized',...
                'Position', [.05 .15 .45 .05]);
            
            self.set_angle_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Set angle',...
                'Units', 'normalized',...
                'Position', [.5 .12 .45 .1],...
                'Callback', @self.set_angle_btn_Callback);
            
            self.att_box = uicontrol(self.runtime_panel, ...
                'Style', 'edit',...
                'String', '0',...
                'Units', 'normalized',...
                'Position', [.05 .35 .45 .05]);
            
            self.set_att_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Set attenuation',...
                'Units', 'normalized',...
                'Position', [.5 .32 .45 .1],...
                'Callback', @self.set_att_btn_Callback);
            
            self.stop_cal_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Stop level cal',...
                'Units', 'normalized',...
                'Position', [.5 .42 .45 .1],...
                'Callback', @self.stop_cal_btn_Callback);
            
            self.start_cal_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Start level cal',...
                'Units', 'normalized',...
                'Position', [.05 .42 .45 .1],...
                'Callback', @self.start_cal_btn_Callback);
            
            self.start_transmission_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Start Periodic',...
                'Units', 'normalized',...
                'Position', [.05 .22 .45 .1],...
                'Callback', @self.start_transmission_btn_Callback);
            
            self.stop_transmission_btn = uicontrol(self.runtime_panel, ...
                'Style', 'pushbutton',...
                'String', 'Stop Periodic',...
                'Units', 'normalized',...
                'Position', [.5 .22 .45 .1],...
                'Callback', @self.stop_transmission_btn_Callback);
            
            static_current_angle_txt = uicontrol(self.runtime_panel,...
                                        'Style', 'text',...
                                        'String', 'Current angle:',...
                                        'HorizontalAlignment', 'left',....
                                        'Units', 'normalized',...
                                        'Position', [0.1,0.9,0.7,0.04]);
                                        
            self.current_angle_txt = uicontrol(self.runtime_panel,...
                                        'Style', 'text',...
                                        'String', '000',...
                                        'HorizontalAlignment', 'left',....
                                        'Units', 'normalized',...
                                        'Position', [0.7,0.9,0.2,0.04]);
                                    
            
                                    
            self.update_from_config();
             
        end
        
        function UpdateGraphics(self)
            
        end
        
        
        function nNodes_box_Callback(self, hObject, eventdata, handles)
            try
                self.config.nNodes = str2double(self.nNodes_box.String);
            catch
                self.config.nNodes = 1;
            end
            self.update_nodeId_list();
        end
        
       
        
        function update_nodeId_list(self)
            self.nodeId_list.String = {1:self.config.nNodes};
            if self.config.nNodes < self.nodeId_list.Value
                self.nodeId_list.Value = 1;
            end
        end
        
        function update_from_config(self)
            self.nodeId_list.Value = self.config.nodeId;
            self.update_nodeId_list(); 
            self.fs_list.Value = find(ismember(self.fs_list.String,num2str(self.config.fs)));
            self.fc_box.String = num2str(self.config.fc);
            self.fm_box.String = num2str(self.config.fm);
            self.baseband_list.Value = self.config.baseband;
            currentIn_idx = find(ismember(self.device_input_list.String, self.config.deviceInName));
            currentOut_idx = find(ismember(self.device_output_list.String, self.config.deviceInName));
            if currentIn_idx > 0
                self.device_input_list.Value = currentIn_idx;
            end
            if currentOut_idx > 0
                self.device_output_list.Value = currentOut_idx;
            end
        end
        
        function update_to_config(self)
            self.config.fs = str2double(self.fs_list.String(self.fs_list.Value));
            self.config.fc = str2double(self.fc_box.String);
            self.config.fm = str2double(self.fm_box.String);
            self.config.nNodes = str2double(self.nNodes_box.String);
            self.config.nodeId = str2double(self.nodeId_list.String(self.nodeId_list.Value));
            self.config.baseband = self.baseband_list.Value;
            self.config.deviceInName = self.device_input_list.String(self.device_input_list.Value);
            self.config.deviceOutName = self.device_output_list.String(self.device_output_list.Value);
            self.saveConfig();
        end
        
        function lock_config_panel(self, enable)
           self.device_input_list.Enable = enable;
           self.device_output_list.Enable = enable;
           self.nNodes_box.Enable = enable;
           self.nodeId_list.Enable = enable;
           self.fs_list.Enable = enable;
           self.fc_box.Enable = enable;
           self.fm_box.Enable = enable;
           self.baseband_list.Enable = enable;
           self.start_btn.Enable = enable;
           
        end
        
        
        
        function start_btn_Callback(self, hObject, eventdata, handles)
            self.update_to_config();
            self.lock_config_panel('off');
            self.status_text.String = '';
            self.running = true;
            self.setupNode();
            self.Play();
        end
        
        function stop_btn_Callback(self, hObject, eventdata, handles)
            self.lock_config_panel('on')
            self.running = false;
        end 
        
        function request_btn_Callback(self, hObject, eventdata, handles)
            self.node.sendRequest();
        end
        
        function cal_btn_Callback(self, hObject, eventdata, handles)
            self.node.calibrate();
        end
        
        function set_angle_btn_Callback(self, hObject, eventdata, handles)
            self.turntable.goToAngle(str2double(self.angle_box.String));
        end
        
        function start_transmission_btn_Callback(self, hObject, eventdata, handles)
            self.node.communicator.startPeriodicTransmission();
        end
        
        function stop_transmission_btn_Callback(self, hObject, eventdata, handles)
            self.node.communicator.stopPeriodicTransmission();
        end
        
        function set_att_btn_Callback(self, hObject, eventdata, handles)
            self.out_factor = 10^(-str2double(self.att_box.String)/10);
        end
        
        function start_cal_btn_Callback(self, hObject, eventdata, handles)
            self.node.level_calibration = 1;
        end
        
        function stop_cal_btn_Callback(self, hObject, eventdata, handles)
            self.node.level_calibration = 0;
        end
        
        
        function createConfig(self)
            
            self.config = struct(...
                'fs',           96E3,...
                'fc',           6E3,...
                'fm',           1E3,...
                'baseband',     1,...
                'nNodes',       3,...
                'nodeId',       1,...
                'deviceInName', 'STUDIO-CAPTURE (core audio)',...
                'deviceOutName','STUDIO-CAPTURE (core audio)');
                
            self.saveConfig();
        end
        
        function saveConfig(self)
            config = self.config;
            save('rtap_config', 'config');
        end
        
        function loadConfig(self)
            try
                config = load('rtap_config');
                self.config = config.config;
            catch
                error('Not able to load config');
            end
        end
        
        function printStatus(self, str)
            self.status_text.String{end+1} = str;
            [idx,~] = size(self.status_text.String);
            if idx > 6
                self.status_text.ListboxTop = idx - 5;
            end
        end
        
    end
    
end
